from ftfy import fix_text
from html import unescape


def repair_html_entities(text):
    before = ''

    while text != before:
        before, text = text, unescape(text)

    return text


def repair_broken_unicode(text):
    return fix_text(text, normalization='NFKD')



import re
from fastbloom_rs import BloomFilter
from unicodedata import normalize


def dedup_lines_bloom(text, just_words=True, zero_digits=True, capacity=100000000, error=0.00001):
    sbf = BloomFilter(
        expected_elements=capacity,
        false_positive_probability=error
    )

    for line in text:
        if not isinstance(line, str):
            raise TypeError('Expected "text" to contain stings, found: {}'.format(type(line)))

        key = line.strip()
        if not key:
            yield line

        key = normalize('NFKD', key)

        if just_words:
            key = ' '.join(re.findall(r'\w+', key))
        if zero_digits:
            key = re.sub(r'\d', '0', key)

        if key in sbf:
            line = ''
        else:
            sbf.add(key)

        yield line



from nlpclean.clean import remove_accent_glyphs
from nlpclean.ddup import dedup_lines_bloom
from nlpclean.html import html_to_article, fragment_to_text
from nlpclean.lang import detect_main_lang
from nlpclean.repair import repair_html_entities, repair_broken_unicode



import re
from bs4 import BeautifulSoup, Comment
from html import unescape
from lxml import etree
from newspaper.cleaners import DocumentCleaner as BaseNewspaperCleaner
from newspaper.configuration import Configuration as NewspaperConfig
from newspaper.extractors import ContentExtractor as NewspaperExtractor
from newspaper.outputformatters import OutputFormatter as NewspaperFormatter

_MEANINGLESS_TAGS = [
    'applet', 'audio', 'blockquote', 'canvas', 'code', 'comment', 'datalist', 'embed', 'figure', 'form', 'frame',
    'frameset', 'iframe', 'img', 'kbd', 'map', 'menu', 'noembed', 'noframes', 'noscript', 'object', 'output',
    'plaintext', 'pre', 'ruby', 'samp', 'script', 'style', 'svg', 'title', 'var', 'video', 'xmp'
]

_BLOCKLEVEL_TAGS = [
    'address', 'article', 'aside', 'blockquote', 'center', 'dd', 'details', 'dialog', 'dir', 'div', 'dl', 'dt',
    'fieldset', 'figcaption', 'figure', 'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header', 'hgroup', 'hr',
    'isindex', 'li', 'main', 'menu', 'nav', 'noframes', 'noscript', 'ol', 'p', 'pre', 'section', 'table', 'td', 'th',
    'tr', 'ul'
]


def html_to_article(content, language):
    content = content.strip()
    if not len(content):
        return ''

    config = NewspaperConfig()
    config.language = language

    doc = config.get_parser().fromstring(content.strip())
    if doc is None:
        return ''

    # Split block-level elements with newlines
    for tag in _BLOCKLEVEL_TAGS:
        if tag in _MEANINGLESS_TAGS:
            continue
        for node in doc.xpath('//{}'.format(tag)):
            node.append(etree.Element('br'))
            node.append(etree.Element('br'))

    # Initial cleanup
    cleaner = _NewspaperCleaner(config)
    doc = cleaner.clean(doc)

    # Best node estimation
    extractor = NewspaperExtractor(config)
    top = extractor.calculate_best_node(doc)
    if top is None:
        del doc, cleaner, extractor
        etree.clear_error_log()

        return ''

    top = extractor.post_cleanup(top)

    # Cleanup dummy nodes used for estimation
    for dummy in top.xpath("//p[@newspaper='dummy']"):
        dummy.getparent().remove(dummy)

    # Custom formatting to avoid unnecessary computations
    formatter = NewspaperFormatter(config)
    formatter.top_node = top
    formatter.remove_negativescores_nodes()
    content = formatter.convert_to_html()
    content = str(content).strip()
    content = unescape(content)

    del doc, top, cleaner, extractor, formatter
    etree.clear_error_log()

    return content


def fragment_to_text(html):
    if not len(html.strip()):
        return ''

    soup = BeautifulSoup(html, 'lxml')
    if soup is None:
        raise ValueError('Can\'t build DOM tree with LXML')

    # Drop comments
    for comment in soup(string=lambda txt: isinstance(txt, Comment)):
        comment.extract()

    # Drop non-meaning tags
    for node in soup(_MEANINGLESS_TAGS):
        node.replace_with(soup.new_tag('br'))

    # Insert linebreaks around block-level tags
    for node in soup(_BLOCKLEVEL_TAGS):
        node.insert_before(soup.new_tag('br'))
        node.insert_before(soup.new_tag('br'))
        node.insert_after(soup.new_tag('br'))
        node.insert_after(soup.new_tag('br'))

    # Remove linebreaks inside text nodes (as browser does)
    for node in soup(string=lambda string: '\n' in string):
        node.string.replace_with(node.string.replace('\n', ' '))

    # Swap html linebreaks to normal ones
    for node in soup('br'):
        node.replace_with('\n')

    # Cleanup final text
    text = soup.getText()
    text = str(text).strip()
    text = re.sub(' {2,}', ' ', text)
    text = text.replace(' \n', '\n').replace('\n ', '\n')
    text = re.sub('\n{3,}', '\n\n', text)

    soup.decompose()
    etree.clear_error_log()
    del soup

    return text


class _NewspaperCleaner(BaseNewspaperCleaner):
    def clean(self, doc_to_clean):
        doc_to_clean = super(_NewspaperCleaner, self).clean(doc_to_clean)

        doc_to_clean = self.div_to_para(doc_to_clean, 'article')

        return doc_to_clean

    def div_to_para(self, doc, dom_type):
        if 'span' == dom_type:
            return doc

        tags = ['a', 'blockquote', 'dl', 'div', 'img', 'ol', 'p', 'pre', 'table', 'ul']
        divs = self.parser.getElementsByTag(doc, tag=dom_type)
        for div in divs:
            items = self.parser.getElementsByTags(div, tags)
            if div is not None and len(items) == 0:
                self.replace_with_para(doc, div)
            elif div is not None:
                self.make_dummy_kid(doc, div)

        return doc

    def make_dummy_kid(self, doc, div):
        text_to_return = []
        replacement_text = []

        if div.text:
            replacement_text.append(div.text.strip())

        for kid in div:
            # The node is a <p> and already has some replacement text
            if self.parser.getTag(kid) == 'p' and len(replacement_text) > 0:
                text_to_return.append(' '.join(replacement_text))
                replacement_text = []
            # The node is a text node
            elif self.parser.getTag(kid) == 'a':
                kid_text = self.parser.outerHtml(kid)
                replacement_text.append(kid_text.strip())

            # The node is a text node
            if kid.tail:
                replacement_text.append(kid.tail.strip())

        # Flush out anything still remaining
        if len(replacement_text):
            text_to_return.append(' '.join(replacement_text))

        for node in text_to_return:
            if not len(node.strip()):
                continue

            node = self.get_flushed_buffer(node, doc)
            if node is None:
                continue

            node.attrib['newspaper'] = 'dummy'
            div.append(node)



import fasttext
import logging
import operator
import os
import requests
import threading
import tqdm
from langid.langid import LanguageIdentifier, model as langid_model
from pycld2 import detect as cld_detect

_langid_model = None
_fasttext_model = None
_init_lock = threading.Lock()


def _download_fasttext():
    remote_url = 'https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin'
    expected_size = 131266198

    model_dir = os.path.expanduser(os.path.join('~', '.nlpclean'))
    model_dir = os.getenv("DATA_DIR") or model_dir
    os.makedirs(model_dir, exist_ok=True)

    model_name = 'lid.176.bin'
    model_path = os.path.join(model_dir, model_name)

    if os.path.exists(model_path) and os.path.getsize(model_path) == expected_size:
        return model_path

    remote_size = int(requests.head(remote_url).headers['Content-Length'])
    if expected_size != remote_size:
        raise AssertionError('Unexpected remote model size')

    bar = tqdm.tqdm(initial=0, total=remote_size, unit='B', unit_scale=True, desc=model_name)
    req = requests.get(remote_url, stream=True)
    with open(model_path, 'wb') as f:
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                bar.update(1024)
    bar.close()

    return model_path


def detect_main_lang(text, warnings=True):
    global _langid_model
    with _init_lock:
        if _langid_model is None:
            _langid_model = LanguageIdentifier.from_modelstring(langid_model, norm_probs=True)

    global _fasttext_model
    with _init_lock:
        if _fasttext_model is None:
            _fasttext_model = fasttext.load_model(_download_fasttext())

    langs = {}

    cld_best_lang, cld_best_score = False, None
    try:
        _, _, scores = cld_detect(text)
        if len(scores):
            cld_best_lang = scores[0][1]
            cld_best_score = scores[0][2] / 100.
        for _, lang, score, _ in scores:
            if lang not in langs:
                langs[lang] = []
            langs[lang].append(score / 100.)
    except Exception as e:
        if warnings:
            logging.warning(e)

    fst_best_lang, fst_best_score = False, None
    try:
        scores = _fasttext_model.predict(text.replace('\n', ' '), 3)
        if len(scores):
            fst_best_lang = scores[0][0].replace('__label__', '')
            fst_best_score = scores[1][0]
        for lang, score in zip(*scores):
            lang = lang.replace('__label__', '')
            if lang not in langs:
                langs[lang] = []
            langs[lang].append(score)
    except Exception as e:
        if warnings:
            logging.warning(e)

    # Fast path
    if cld_best_lang == fst_best_lang and cld_best_lang and 'un' != cld_best_lang:
        score = (cld_best_score + fst_best_score) / 2.
        sure = score > 0.75
        return cld_best_lang, sure, score

    try:
        lang, score = _langid_model.classify(text)
        if lang not in langs:
            langs[lang] = []
        langs[lang].append(score)
    except Exception as e:
        if warnings:
            logging.warning(e)

    if not langs:
        return '', False, 0.0

    scored = map(lambda x: (x[0], sum(x[1]) / len(x[1])), langs.items())
    lang, score = max(scored, key=operator.itemgetter(1))
    sure = len(langs[lang]) == 1 or len(langs[lang]) > 1 and score >= 0.75

    return lang, sure, score
